# Copyright: Martin Matusiak <numerodix@gmail.com>

from __future__ import absolute_import

import sys


LINEWIDTH = 78

def write(s):
    sys.stdout.write(s)
    sys.stdout.flush()

def writeln(s):
    clear_line()
    write(s + '\n')

def clear_line():
    write(LINEWIDTH * ' ' + '\r')



# Copyright: Martin Matusiak <numerodix@gmail.com>

from __future__ import absolute_import

import itertools
import re
import string


class DigitString(object):
    def __init__(self, fp, prefix='', postfix=''):
        self.prefix = prefix
        self.postfix = postfix

        digit_runs = re.finditer("([0-9]+)", fp)
        nondigit_runs = re.finditer("([^0-9]+)", fp)

        digit_spans = [m.span() for m in digit_runs if digit_runs]
        nondigit_spans = [m.span() for m in nondigit_runs if nondigit_runs]

        if nondigit_spans and nondigit_spans[0][0] != 0:
            nondigit_spans = [(0, 0)] + nondigit_spans

        def take_slice(tup):
            x, y = tup
            return fp[x:y]

        self.numbers = map(take_slice, digit_spans)
        self.chars = map(take_slice, nondigit_spans)

    def has_digits(self):
        return len(self.numbers) > 0

    def process_field_number(self, field):
        assert(field != 0)
        field = field - 1 if field > 0 else field
        return field

    def get_field_count(self):
        return len(self.numbers)

    def get_field(self, field):
        field = self.process_field_number(field)
        try:
            return self.numbers[field]
        except IndexError:
            return ""

    def set_field(self, field, number):
        field = self.process_field_number(field)
        assert(type(number) == str)
        assert(re.match('^[0-9]+$', number))
        try:
            self.numbers[field] = number
        except IndexError:
            pass

    def set_field_width(self, field, width):
        val = self.get_field(field)
        if val:
            val = string.zfill(str(int(val)), width)
            self.set_field(field, val)

    def get_string(self):
        s = ''
        for (a, b) in itertools.izip_longest(self.chars,
                                             self.numbers, fillvalue=''):
            s += a + b
        return self.prefix, s, self.postfix



# Copyright: Martin Matusiak <numerodix@gmail.com>

from __future__ import absolute_import

import os
import re

from nametrans import callbacks
from nametrans.callbacks import RenameException

EXCEPTION_LIST = (RenameException, OSError)


class Fs(object):
    fs_case_sensitive = None

    @classmethod
    def detect_fs_case_sensitivity(cls, path):
        '''Detect the case sensitivity of a filesystem. Given a path, find a
        file in that directory and check os.path.exists in lowercase and
        uppercase. Should be a reliable test for the specific path, but does
        not remedy crossing filesystem boundaries with mounted
        filesysystems.'''
        # find a file to test on
        fp = None
        if os.path.isdir(path):
            fps = os.listdir(path)
            if fps:
                fp = fps[0]

        if not fp:
            fp = os.path.basename(path)

        cls.fs_case_sensitive = not (os.path.exists(fp.lower()) and
                                     os.path.exists(fp.upper()))

    @classmethod
    def find(cls, path, rec=False):
        cls.detect_fs_case_sensitivity(path)

        def remove_basepath(p):
            rx = '^' + re.escape(path) + '(?:' + re.escape(os.sep) + ')?'
            return re.sub(rx, '', p)

        def write_progress(p):
            p = remove_basepath(p)
            callbacks.progress("Scanning:", (p and p or '.'))

        fs = []
        if not rec:
            write_progress(path)
            fs = os.listdir(path)
        else:
            for p, dirs, files in os.walk(path):
                write_progress(p)
                for fp in dirs + files:
                    fp = os.path.join(p, fp)
                    fs.append(fp)
        fs = map(remove_basepath, fs)
        return sorted(fs)

    @classmethod
    def string_normalize_filepath(cls, fp):
        if not cls.fs_case_sensitive:
            fp = fp.lower()
        return fp

    @classmethod
    def string_is_same_file(cls, f, g):
        """Check if files are the same on disk"""
        return (cls.string_normalize_filepath(f) ==
                cls.string_normalize_filepath(g))

    @classmethod
    def io_invalid_rename(cls, f, g):
        """Handle rename on case insensitive fs, test not only for file exists,
        but also that it's the same file"""
        return os.path.exists(g) and not cls.string_is_same_file(f, g)


    @classmethod
    def do_rename_with_temp_exc(cls, func, f, g):
        """Rename f -> g, but using t as tempfile. If writing to g fails,
        will attempt to rollback f <- t."""
        t = f + '.tmp'
        while os.path.exists(t):
            t += 'z'

        # f -> t
        try:
            func(f, t)
        except OSError:
            raise OSError("Failed to create tempfile for rename: %s" % t)

        # t -> g
        try:
            func(t, g)
        except OSError:
            try:
                func(t, f)
            except OSError:
                raise OSError("Failed to rollback rename: %s <- %s" % (f, t))
            raise OSError("Failed rename %s -> %s" % (f, g))

    @classmethod
    def do_rename_exc(cls, f, g):
        """Attempts to detect an overwrite, throwing RenameException.
        If detection fails, will throw OSError."""
        if cls.io_invalid_rename(f, g):
            raise RenameException("Target exists: %s" % g)
        else:
            cls.do_rename_with_temp_exc(os.renames, f, g)

    @classmethod
    def do_renamedir(cls, f, g):
        if not os.path.exists(g) or cls.string_is_same_file(f, g):
            try:
                cls.do_rename_with_temp_exc(os.rename, f, g)
            except EXCEPTION_LIST as e:
                callbacks.error_handler(e)
        else:
            for fp in os.listdir(f):
                try:
                    cls.do_rename_exc(os.path.join(f, fp), os.path.join(g, fp))
                except EXCEPTION_LIST as e:
                    callbacks.error_handler(e)

    @classmethod
    def io_set_actual_path(cls, filepath):
        """Fix a filepath that has the wrong case on the fs by renaming
        its parts directory by directory"""
        parts = filepath.split(os.sep)
        for (i, part) in enumerate(parts):
            prefix = os.sep.join(parts[:i]) if i > 0 else '.'
            fps = os.listdir(prefix)
            for fp in fps:
                if part.lower() == fp.lower() and not part == fp:
                    prefix = '' if prefix == '.' else prefix
                    fp_fs = os.path.join(prefix, fp)
                    fp_target = os.path.join(prefix, part)
                    cls.do_renamedir(fp_fs, fp_target)
                    break

    @classmethod
    def do_renames(cls, lst):
        for (f, g) in lst:
            try:
                cls.do_rename_exc(f, g)
            except EXCEPTION_LIST as e:
                callbacks.error_handler(e)

        # another pass on the dirs for case fixes
        dirlist = {}
        for (f, g) in lst:
            d = os.path.dirname(g)
            if d and d not in dirlist and os.path.exists(d):
                dirlist[d] = None
                cls.io_set_actual_path(d)



# Copyright: Martin Matusiak <numerodix@gmail.com>

from __future__ import absolute_import

import re


class FilepathTransformer(object):
    @classmethod
    def to_unicode(self, s):
        us = s.decode('utf-8', 'ignore')
        return us

    @classmethod
    def from_unicode(self, us):
        s = us.encode('utf-8', 'ignore')
        return s

    @classmethod
    def by_regex(cls, rx_from, rx_to, s):
        return re.sub(rx_from, rx_to, s)

    @classmethod
    def capitalize(cls, s):
        us = cls.to_unicode(s)
        cap = lambda m: m.group(1).upper() + m.group(2).lower()
        us = re.sub("(?u)(?<![0-9\w'])(\w)([\w']*)", cap, us)
        s = cls.from_unicode(us)
        return s

    @classmethod
    def make_lowercase(cls, s):
        us = cls.to_unicode(s)
        tolower = lambda m: m.group(1).lower()
        us = re.sub('(?u)([\w]*)', tolower, us)
        s = cls.from_unicode(us)
        return s

    @classmethod
    def make_spaces_underscores(cls, s):
        s = re.sub(' ', '_', s)
        return s

    @classmethod
    def do_trim(cls, s):
        # check endpoints
        s = cls.by_regex('^([ ]|-)*', '', s)
        s = cls.by_regex('([ ]|-)*$', '', s)
        return s

    @classmethod
    def make_neat(cls, s):
        # too many hyphens and underscores
        s = cls.by_regex('_{2,}', '-', s)
        s = cls.by_regex('-{2,}', '-', s)
        s = cls.by_regex('-[ ]+-', '-', s)
        # junk-y chars past the start of the string
        s = cls.by_regex('(?<!^)\.', ' ', s)
        s = cls.by_regex('_', ' ', s)
        s = cls.by_regex('#', ' ', s)
        s = cls.by_regex(':', ' ', s)
        # let's have spaces around hyphen
        s = cls.by_regex('(?<!\s)-', ' -', s)
        s = cls.by_regex('-(?!\s)', '- ', s)
        s = cls.by_regex('(?<!\s)[+]', ' +', s)
        s = cls.by_regex('[+](?!\s)', '+ ', s)
        # empty brackets
        s = cls.by_regex('\[ *?\]', ' ', s)
        s = cls.by_regex('\( *?\)', ' ', s)
        # normalize spaces
        s = cls.by_regex('[ ]{2,}', ' ', s)
        s = cls.do_trim(s)
        return s

    @classmethod
    def make_neater(cls, s):
        # bracket-y junk
        s = cls.by_regex('\[.*?\]', ' ', s)
        s = cls.by_regex('\(.*?\)', ' ', s)
        s = cls.do_trim(s)
        return s



#!/usr/bin/env python
#
# Copyright: Martin Matusiak <numerodix@gmail.com>

from __future__ import absolute_import

import os
import re
import sre_constants
import string
import sys

import ansicolor

from nametrans import callbacks
from nametrans import io
from nametrans import nametransformer
from nametrans.fs import Fs
from nametrans.nametransformer import NameTransformer


class Program(object):
    def __init__(self, options):
        self.options = options
        self.nameTransformer = NameTransformer(options)

    def validate_options(self):
        try:
            re.compile(self.options.s_from)
            re.compile(self.options.s_to)
            return True
        except (sre_constants.error, re.error) as e:
            re_exc = callbacks.RegularExpressionError(*e.args)
            callbacks.error_handler(re_exc)

    def display_transforms_and_prompt(self, items):
        clashes = 0
        arrow = "->"
        prefix = " * "
        linewidth = 78
        spacing = 2

        def get_slot(linewidth, arrow, prefix, spacing):
            return (linewidth - len(arrow) - prefix - spacing) / 2

        slotlong = get_slot(linewidth, arrow, len(prefix), spacing)
        longest_l = max(map(lambda item: len(item.f), items))
        longest = max(longest_l, max(map(lambda item: len(item.g), items)))
        slot = longest
        slot_l = longest_l
        if longest > slotlong:
            slot = get_slot(linewidth, arrow, len(prefix), spacing)
            slot_l = slot

        for item in items:
            arrow_fmt = ansicolor.yellow(arrow)
            prefix_fmt = ansicolor.green(prefix)
            f_fmt, g_fmt = ansicolor.colordiff(item.f, item.g)
            if item.invalid:
                clashes += 1
                g_fmt = ansicolor.red(item.g)
            if len(item.f) <= slot and len(item.g) <= slot:
                f_fmt = ansicolor.justify_formatted(f_fmt,
                                                    string.ljust, slot_l)
                io.writeln("%s%s %s %s" %
                           (prefix_fmt, f_fmt, arrow_fmt, g_fmt))
            else:
                io.writeln("%s%s\n%s %s" %
                           (prefix_fmt, f_fmt, arrow_fmt, g_fmt))

        s_files = "%s files" % len(items)
        prompt = "Rename %s? [y/N] " % s_files
        if clashes:
            prompt = ("%s clash(es) exist, rename %s? [y/N] " %
                      (clashes, s_files))

        sys.stdout.write(prompt)
        inp = raw_input()

        return inp == "y"

    def perform_renames(self, items):
        callbacks.progress("Performing renames...")

        pairs = map(lambda it: (it.f, it.g), items)
        Fs.do_renames(pairs)

        io.clear_line()

    def run(self):
        if self.validate_options():
            items = self.nameTransformer.scan_fs()
            items = self.nameTransformer.process_items(items)
            if items and self.display_transforms_and_prompt(items):
                self.perform_renames(items)


def main():
    options, args, parser = nametransformer.get_opt_parse(sys.argv)

    # options that don't need from/to patterns
    if not args and not any([
        options.flag_capitalize,
        options.flag_lowercase,
        options.flag_neat,
        options.flag_neater,
        options.flag_underscore,
        options.flag_dirname,
        options.renseq,
        options.flag_flatten,
    ]):
        parser.print_help()
        sys.exit(0)

    if not os.path.exists(options.path):
        io.writeln("Invalid path: %s" % options.path)
        sys.exit(1)
    else:
        os.chdir(options.path)

    Program(options).run()


if __name__ == '__main__':
    main()



# Copyright: Martin Matusiak <numerodix@gmail.com>

from __future__ import absolute_import

from nametrans import io


class RenameException(Exception):
    pass

class RegularExpressionError(Exception):
    pass


def error_handler(exc):
    msg = ' '.join(exc.args)
    io.writeln("%s: %s" % (exc.__class__.__name__, msg))

def _get_progress_line(*args):
    action, arg = args[0], ''
    if len(args) > 1:
        arg = " ".join(args[1:])

    def get_line(action, arg):
        line = "%s %s" % (action, arg)
        return line

    linelen = io.LINEWIDTH
    space = 1
    padding = 3

    line = get_line(action, arg)
    if len(line) > linelen:
        width = linelen - len(action) - space - padding
        arg = '.' * padding + arg[-width:]
        line = get_line(action, arg)

    line = line.ljust(linelen)
    return line

def progress(*args):
    line = _get_progress_line(*args)
    io.write(line + '\r')



from setuptools import find_packages
from setuptools import setup

import nametrans


setup(
    name='nametrans',
    version=nametrans.__version__,
    description='Rename files with regex search/replace semantics',
    author='Martin Matusiak',
    author_email='numerodix@gmail.com',
    url='https://github.com/numerodix/nametrans',

    packages=find_packages('.'),
    package_dir={'': '.'},

    install_requires=[
        'ansicolor',
    ],

    entry_points={
        "console_scripts": [
            "nametrans = nametrans.main:main",
        ]
    },

    # don't install as zipped egg
    zip_safe=False,

    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
)



# Copyright: Martin Matusiak <numerodix@gmail.com>

from __future__ import absolute_import

import os
import re
import string
from optparse import OptionParser

from nametrans import callbacks
from nametrans import io
from nametrans.digitstring import DigitString
from nametrans.filepathtrans import FilepathTransformer
from nametrans.fs import Fs


class FilePath(object):
    def __init__(self, fp):
        self.f, self.g = fp, fp
        self.invalid = False

class NameTransformer(object):
    def __init__(self, options):
        if options.flag_neat or options.flag_neater:
            options.flag_root = True
            if not options.flag_lowercase:
                options.flag_capitalize = True
        if options.flag_dirname:
            options.flag_root = True
            options.flag_filesonly = True
        if options.renseq:
            options.flag_root = True
            options.flag_filesonly = True

        self.options = options

    def scan_fs(self):
        fps = Fs.find(self.options.path,
                      rec=any([self.options.flag_recursive,
                               self.options.flag_flatten]))

        callbacks.progress("Separating files and directories...")
        file_items = filter(os.path.isfile, fps)
        dir_items = filter(os.path.isdir, fps)

        items = file_items
        if self.options.flag_dirsonly or not items:
            items = dir_items

        return map(FilePath, items)

    def get_patterns(self):
        s_from = self.options.s_from
        s_to = self.options.s_to
        if s_from:
            if self.options.flag_literal:
                s_from = re.escape(s_from)
                s_to = re.sub('\\g', '\\\\g', s_to)  # escape \g
            if self.options.flag_ignorecase:
                s_from = '(?i)' + s_from
        return s_from, s_to

    def split_filepath(self, fp):
        path, root, ext = '', fp, ''
        if self.options.flag_filesonly and self.options.flag_root:
            path, name = os.path.split(fp)
            root, ext = os.path.splitext(name)
        elif self.options.flag_filesonly:
            path, root = os.path.split(fp)
        elif self.options.flag_root:
            root, ext = os.path.splitext(fp)
        return path, root, ext

    def index_items_by_dir(self, items):
        dirindex = {}
        for item in items:
            d = os.path.dirname(item.g) or '.'
            if d not in dirindex:
                dirindex[d] = [item]
            else:
                dirindex[d].append(item)
        return dirindex

    def apply_flatten(self, items):
        for item in items:
            item.g = re.sub(re.escape(os.sep), ' - ', item.g)
        return items

    def apply_dirname(self, items):
        dirindex = self.index_items_by_dir(items)

        items = []
        for (d, its) in dirindex.items():
            if d == '.':
                d = os.path.basename(os.path.abspath(d))
            w = len(str(len(its)))
            for (i, item) in enumerate(sorted(its)):
                path, _, ext = self.split_filepath(item.g)
                root = "%s %s" % (os.path.basename(d), string.zfill(i + 1, w))
                item.g = os.path.join(path, root) + ext
                items.append(item)

        return items

    @classmethod
    def parse_renseq_args(cls, s):
        is_set, field, width = False, 0, 0
        if s:
            parts = s.split(':')

            width = "0"
            if len(parts) == 1:
                field = parts[0]
            if len(parts) == 2:
                field = parts[0]
                width = parts[1] or "0"

            if field:
                field = int(field)
            width = int(width)

            is_set = True

        return is_set, field, width

    def apply_renseq(self, arg_field, arg_width, items):
        dirindex = self.index_items_by_dir(items)
        items = []
        for (d, its) in dirindex.items():
            dstrings = []
            maxfields = 0
            for item in its:
                path, filename, ext = self.split_filepath(item.g)
                digitstring = DigitString(filename, prefix=path, postfix=ext)
                if digitstring.has_digits():
                    dstrings.append((item, digitstring))
                    maxfields = max(maxfields, digitstring.get_field_count())

            fields = range(1, maxfields + 1)
            if arg_field:
                fields = [arg_field]

            for field in fields:
                # find width from names
                width = arg_width if arg_width else 0
                if not width:
                    maxlen = 0
                    for (item, digitstring) in dstrings:
                        v = digitstring.get_field(field)
                        if v:
                            maxlen = max(maxlen, len(str(int(v))))
                    width = maxlen

                for (item, digitstring) in dstrings:
                    digitstring.set_field_width(field, width)

            for (item, digitstring) in dstrings:
                path, filename, ext = digitstring.get_string()
                item.g = os.path.join(path, filename) + ext
                items.append(item)

        return items

    def compute_transforms(self, items):
        if self.options.flag_flatten:
            items = self.apply_flatten(items)
        if self.options.flag_dirname:
            items = self.apply_dirname(items)
        if self.options.renseq:
            items = self.apply_renseq(self.options.renseq_field,
                                      self.options.renseq_width,
                                      items)

        for item in items:
            path, t, ext = self.split_filepath(item.g)
            s_from, s_to = self.get_patterns()

            t = FilepathTransformer.by_regex(s_from, s_to, t)
            if self.options.flag_neat or self.options.flag_neater:
                ext = FilepathTransformer.make_lowercase(ext)
                t = FilepathTransformer.make_neat(t)
                if self.options.flag_neater:
                    t = FilepathTransformer.make_neater(t)
            if self.options.flag_capitalize:
                t = FilepathTransformer.capitalize(t)
            if self.options.flag_lowercase:
                t = FilepathTransformer.make_lowercase(t)
            if self.options.flag_underscore:
                t = FilepathTransformer.make_spaces_underscores(t)

            item.g = os.path.join(path, t + ext)
        return items

    def compute_clashes(self, items):
        index = {}
        for item in items:
            fp = Fs.string_normalize_filepath(item.g)
            if fp not in index:
                index[fp] = item
            else:
                item.invalid = True
                index[fp].invalid = True
            if Fs.io_invalid_rename(item.f, item.g):
                item.invalid = True
        return items

    def process_items(self, items):
        callbacks.progress("Computing renames...")
        items = self.compute_transforms(items)

        # no change in name
        if not self.options.renseq:
            items = filter(lambda item: item.f != item.g, items)
        # rename to empty
        items = filter(lambda item: item.g != '', items)

        callbacks.progress("Checking for clashes...")
        items = self.compute_clashes(items)

        items.sort(key=lambda item: (item.g.lower(), item.f.lower()))

        io.clear_line()
        return items



def get_opt_parse(argv):
    usage = 'Usage:  %s [options] "<from>" "<to>"\n' % argv[0]

    usage += '\n$ %s "apple" "orange"' % argv[0]
    usage += '\n * I like apple.jpg -> I like orange.jpg'
    usage += '\n * pineapple.jpg    -> pineorange.jpg'

    parser = OptionParser(usage=usage)
    parser.add_option("--path", help="Run on this path",
                      dest="path", action="store")
    parser.add_option("-r", help="Apply recursively",
                      dest="flag_recursive", action="store_true")
    parser.add_option("--dirs", help="Apply rename to directories, not files",
                      dest="flag_dirsonly", action="store_true")
    parser.add_option("--files", help="Apply rename only to files",
                      dest="flag_filesonly", action="store_true")
    parser.add_option("--lit",
                      help="Treat patterns as literal, not regular expression",
                      dest="flag_literal", action="store_true")
    parser.add_option("-i", help="Apply pattern ignoring case",
                      dest="flag_ignorecase", action="store_true")
    parser.add_option("--root",
                      help="Apply rename to root only (not extension)",
                      dest="flag_root", action="store_true")
    parser.add_option("--cap", help="Capitalize",
                      dest="flag_capitalize", action="store_true")
    parser.add_option("--lower", help="Make lowercase",
                      dest="flag_lowercase", action="store_true")
    parser.add_option("--neat", help="Make neat",
                      dest="flag_neat", action="store_true")
    parser.add_option("--neater", help="Remove more junk than regular neat",
                      dest="flag_neater", action="store_true")
    parser.add_option("--under", help="Use underscores for spaces",
                      dest="flag_underscore", action="store_true")
    parser.add_option("--dirname",
                      help="Use the current directory name as filename",
                      dest="flag_dirname", action="store_true")
    parser.add_option("--renseq", help="Change width of numbers in names",
                      dest="renseq", action="store", metavar="field:width")
    parser.add_option("--flatten",
                      help="Flatten directory tree to flat directory",
                      dest="flag_flatten", action="store_true")
    (options, args) = parser.parse_args(argv[1:])

    tup = NameTransformer.parse_renseq_args(options.renseq)
    options.renseq, options.renseq_field, options.renseq_width = tup

    path = options.path
    if not options.path:
        path = os.getcwd()
    options.path = os.path.abspath(path)

    options.s_from, options.s_to = '', ''
    try:
        argsdupe = args[:]
        options.s_from = argsdupe.pop(0)
        options.s_to = argsdupe.pop(0)
    except IndexError:
        pass

    return options, args, parser
